use {
    crate::Node,
    std::{
        io::{self, BufRead, Write},
        string::ToString,
    },
};

#[derive(Debug, Default)]
enum Mode {
    #[default]
    Normal,
    Preformatted,
}

#[derive(Debug)]
pub struct Parser<B: BufRead, W: Write> {
    reader: B,
    writer: Option<W>,
    mode: Mode,
}

impl<B: BufRead, W: Write> Iterator for Parser<B, W> {
    type Item = io::Result<Node>;

    fn next(&mut self) -> Option<Self::Item> {
        let mut s = String::new();
        let node = match self.reader.read_line(&mut s) {
            Ok(0) => None,
            Err(e) => Some(Err(e)),
            Ok(_) if self.is_preformatted() => Some(Ok(self.parse_preformatted(&s))),
            Ok(_) => Some(Ok(self.parse_normal(&s))),
        };
        if let Some(ref mut w) = self.writer {
            if let Some(Ok(_)) = node {
                // We throw away the possible error here if writing to our cache
                // fails. TODO: Log this somehow
                let _s = writeln!(w, "{s}");
            }
        }
        node
    }
}

impl<B: BufRead, W: Write> Parser<B, W> {
    pub fn new(reader: B, writer: Option<W>) -> Self {
        Self {
            reader,
            writer,
            mode: Mode::default(),
        }
    }

    fn is_preformatted(&self) -> bool {
        matches!(self.mode, Mode::Preformatted)
    }

    fn parse_normal(&mut self, s: &str) -> Node {
        if let Some(s) = s.strip_prefix("```").map(str::trim) {
            self.mode = Mode::Preformatted;
            Self::parse_pre_alt(s)
        } else if let Some(s) = s.strip_prefix("=>").map(str::trim) {
            Self::parse_link(s)
        } else if let Some(s) = s.strip_prefix("=:").map(str::trim) {
            Self::parse_prompt(s)
        } else if let Some(s) = s
            .strip_prefix("###")
            .map(str::trim)
            .map(ToString::to_string)
        {
            Node::H3(s)
        } else if let Some(s) = s.strip_prefix("##").map(str::trim).map(ToString::to_string) {
            Node::H2(s)
        } else if let Some(s) = s.strip_prefix('#').map(str::trim).map(ToString::to_string) {
            Node::H1(s)
        } else if let Some(s) = s.strip_prefix('*').map(str::trim).map(ToString::to_string) {
            Node::ListItem(s)
        } else if let Some(s) = s.strip_prefix('>').map(str::trim).map(ToString::to_string) {
            Node::Quote(s)
        } else {
            Node::Text(s.to_string())
        }
    }

    fn parse_preformatted(&mut self, s: &str) -> Node {
        if s.starts_with("```") {
            self.mode = Mode::Normal;
            Node::PreEnd
        } else {
            Node::PreBody(s.to_string())
        }
    }

    fn parse_link(s: &str) -> Node {
        let (url, display) = s
            .split_once(|x: char| x.is_ascii_whitespace())
            .map(|(u, d)| (u.trim().to_string(), Some(d.trim().to_string())))
            .unwrap_or((s.to_string(), None));
        Node::Link { url, display }
    }

    fn parse_prompt(s: &str) -> Node {
        let (url, display) = s
            .split_once(|x: char| x.is_ascii_whitespace())
            .map(|(u, d)| (u.trim().to_string(), Some(d.trim().to_string())))
            .unwrap_or((s.to_string(), None));
        Node::Prompt { url, display }
    }

    fn parse_pre_alt(s: &str) -> Node {
        let alt = if s.is_empty() {
            None
        } else {
            Some(s.to_string())
        };
        Node::PreStart { alt }
    }
}
